/*--------------------------------------------------------------------*/
/*--- Callgrind                                                    ---*/
/*---                                                 ct_include.h ---*/
/*--- (C) 2004, Josef Weidendorfer                                 ---*/
/*--------------------------------------------------------------------*/


#ifndef CT_INCLUDE
#define CT_INCLUDE

#include "vg_skin.h"
#include "ct_events.h"

/*------------------------------------------------------------*/
/*--- Calltree compile options                            --- */
/*------------------------------------------------------------*/

/* Enable debug output */
#define CT_ENABLE_DEBUG 1

/* Enable experimental features? */
#define CT_EXPERIMENTAL 1

/* Enable data collection ? */
#define CT_DATA 0

/* Enable hardware prefetch simulation ? */
#define CT_HWPREFETCH 1

/* Enable assertions for cc/fcc operations */
#define ENABLE_CC_ASSERTIONS 0

/* Set to 1 if you want full sanity checks for JCC */
#define JCC_CHECK 0



/*------------------------------------------------------------*/
/*--- Command line args                                    ---*/
/*------------------------------------------------------------*/

/* Dump format options */
extern Char* clo_filename_base;
extern Bool clo_separate_dumps;
extern Bool clo_compress_strings;
extern Bool clo_compress_events;
extern Bool clo_compress_pos;
extern Bool clo_mangle_names;
extern Bool clo_compress_mangled;
extern Bool clo_dump_line;
extern Bool clo_dump_instr;
extern Bool clo_dump_bb;
extern Bool clo_dump_bbs;
extern Bool clo_dump_skipped;

/* Dump generation options */
extern Int clo_dump_every_bb;

/* Collection options */
extern Bool clo_separate_threads;
extern Bool clo_collect_data;
extern Bool clo_collect_atstart;
extern Bool clo_collect_jumps;
extern Bool clo_skip_plt;
extern Int  clo_separate_callers;
extern Int  clo_separate_recursions;
extern Bool clo_skip_direct_recursion;

/* Simulation options */
extern Bool clo_simulate_cache;
extern Bool clo_simulate_hwpref;
extern Bool clo_instrument_atstart;


#define DEFAULT_DUMPNAME    "callgrind.out"
#define DEFAULT_COMMANDNAME "callgrind.cmd"
#define DEFAULT_RESULTNAME  "callgrind.res"
#define DEFAULT_INFONAME    "/tmp/callgrind.info"

/*------------------------------------------------------------*/
/*--- Constants                                            ---*/
/*------------------------------------------------------------*/


/* According to IA-32 Intel Architecture Software Developer's Manual: Vol 2 */
#define MAX_x86_INSTR_SIZE              16

/* Minimum cache line size allowed */
#define MIN_LINE_SIZE   16

/* Size of various buffers used for storing strings */
#define FILENAME_LEN                    256
#define FN_NAME_LEN                    4096 /* for C++ code :-) */
#define OBJ_NAME_LEN                    256
#define BUF_LEN                         512
#define COMMIFY_BUF_LEN                 128
#define RESULTS_BUF_LEN                 128
#define LINE_BUF_LEN                     64



/*------------------------------------------------------------*/
/*--- Structure declarations                               ---*/
/*------------------------------------------------------------*/

typedef struct _Context     Context;
typedef struct _CC          CC;
typedef struct _BB          BB;
typedef struct _Skipped     Skipped;
typedef struct _BBCC        BBCC;
typedef struct _jCC         jCC;
typedef struct _fCC         fCC;
typedef struct _fn_node     fn_node;
typedef struct _file_node   file_node;
typedef struct _obj_node    obj_node;
typedef struct _fn_config   fn_config;
typedef struct _fn_info     fn_info;
typedef struct _call_entry  call_entry;
typedef struct _thread_cxtinfo thread_cxtinfo;
typedef struct _thread_info thread_info;
typedef struct _event_type  event_type;
typedef struct _stat_info   stat_info;

/* For cache simulation */
typedef struct {
    int size;       /* bytes */ 
    int assoc;
    int line_size;  /* bytes */ 
} cache_t;


/* Costs of event sets. Aliases to arrays of 64-bit values */
typedef ULong* CallCost;
typedef ULong* LineCost;


/* A simple cost metric with 2 dependencies:
 * Cost m1 always happens with a, and m2 always happens with m1.
 * E.g. a 2nd level cache miss can only happen with a 1st level miss
 *      and a data access
 *
 * This semantic is used in operations add_cc/add_diff_cc.
 */
struct _CC {
   ULong a;
   ULong m1;
   ULong m2;
};


/* a full cost center - with instruction, data read and write costs.
 * No further assertions...
 * (with REP an instruction can produce data > instruction accesses)
 */
struct _fCC {
    CC Ir;
    CC Dr;
    CC Dw;
};




/* JmpCall cost center
 * for subroutine call (from->bb->jmp_addr => to->bb->addr)
 *
 * Each BB has at most one CALL instruction. The list of JCC from
 * this call is a pointer to the list head (stored in BBCC), and
 * <next_from> in the JCC struct.
 *
 * For fast lookup, JCCs are reachable with a hash table, keyed by
 * the (from_bbcc,to) pair. <next_hash> is used for the JCC chain
 * of one hash table entry.
 *
 * Cost <sum> holds event counts for already returned executions.
 * <last> are the event counters at last enter of the subroutine.
 * <sum> is updated on returning from the subroutine by
 * adding the diff of <last> and current event counters to <sum>.
 *
 * After updating, <last> is set to current event counters. Thus,
 * events are not counted twice for recursive calls (TODO: True?)
 */

#define JmpNone -1 
#define JmpCond -2

struct _jCC {
    Int  jmpkind;     /* JmpCall, JmpBoring, JmpCond */
    jCC* next_hash;   /* for hash entry chain */
    jCC* next_from;   /* next JCC from a BBCC */
    BBCC *from, *to;  /* call arc from/to this BBCC */

    ULong call_counter; /* no wraparound with 64 bit */
    fCC sum;
};


/**
 * An instrumented basic block (BB).
 *
 * BBs are put into a resizable hash to allow for fast detection if a
 * BB is to be retranslated but cost info is already available.
 * The key for a BB is a (object, offset) tupel making it independent
 * from possibly multiple mappings of the same ELF object.
 *
 * At the beginning of each instrumented BB,
 * a call to setup_bbcc(), specifying a pointer to the
 * according BB structure, is added.
 *
 * As cost of a BB has to be distinguished depending on the context,
 * multiple cost centers for one BB (struct BBCC) exist and the according
 * BBCC is set by setup_bbcc.
 */
struct _BB {
  obj_node*  obj;         /* ELF object of BB */
  UInt       offset;      /* file offset of BB in object file */

  Addr       addr;        /* TODO: remove */
  UInt       size;        /* length of BB in bytes */
  Addr       jmp_addr;    /* address of the 1st jmp instruction in this BB */
  fn_node*   fn;          /* debug info for this BB */
  Int        line;        /* line debug info */
  Bool       is_entry;    /* True if this BB is a function entry */
        
  VgSectKind sect_kind;  /* section of this BB, e.g. PLT */
  UInt       array_size; /* byte-size of variable length array in BBCCs */
  UInt       instr_count;
  
  BBCC*      bbcc_list;  /* BBCCs for same BB (see next_bbcc in BBCC) */
  BBCC*      last_bbcc;  /* Temporary: Cached for faster access (LRU) */
  BB*        next;       /* chaining for a hash entry */

  EventList  info;       /* info on instruction sizes and needed costs */
};



/**
 * Execution context
 *
 * Basic blocks are always executed in a context.
 * A execution context is a list of function nodes.
 * Recursion levels are handled for active function and thread.
 *
 * To get a unique number for a full execution context, use
 *  rec_index = min(<fn->rec_separation>,<active>) - 1;
 *  unique_no = <number> + rec_index
 *
 * For each Context, recursion index and BB, there can be a BBCC.
 */
struct _Context {
    UInt size;        // number of function dependencies
    UInt base_number; // for context compression & dump array
    Context* next;    // entry chaining for hash
    UInt hash;        // for faster lookup...
    fn_node* fn[0];
};

/*
 * Skipped cost item
 *
 * Cost for skipped functions is collected in this struct,
 * allocated lazy when a call is done to a skipped function from
 * <fromBBCC>. We note the skipped context.
 * As many functions could be skipped in a call chain,
 * skipped cost items form a list. 
 */
struct _Skipped {
    fn_node* fn;
    BBCC* from;
    Skipped* next;
    fCC fcc;
};

/*
 * Basic Block Cost Center
 *
 * On demand, multiple BBCCs will be created for the same BB
 * dependend on command line options and:
 * - current function (it's possible that a BB is executed in the
 *   context of different functions, e.g. in manual assembler/PLT)
 * - current thread ID
 * - position where current function is called from
 * - recursion level of current function
 *
 * The cost centres for the instructions of a basic block are
 * stored in a contiguous array.
 * They are distinguishable by their tag field.
 */
struct _BBCC {
    BB*      bb;           /* BB for this cost center */

    Context* cxt;          /* execution context of this BBCC */
    ThreadId tid;          /* only for assertion check purpose */
    UInt     rec_index;    /* Recursion index in rec->bbcc for this bbcc */

    ULong    exe_counter;  /* execution counter for BB in this context */
    ULong    ret_counter;  /* how often returned from jccs of this bbcc */

    BBCC**   rec_array;    /* Array of pointers to recursion BBCCs */
    BBCC*    next_bbcc;    /* Chain of BBCCs for same BB */
    BBCC*    lru_next_bbcc; /* BBCC executed next the last time */
    
    jCC*     lru_from_jcc; /* Temporary: Cached for faster access (LRU) */
    jCC*     lru_to_jcc;   /* Temporary: Cached for faster access (LRU) */
    jCC*     jcc_list;     /* list of arcs called from jmp_addr */
    Skipped* skipped;      /* cost for skipped functions called from 
			    * jmp_addr. Allocated lazy */

    BBCC*    next1;         /* entry chain in 1st hash */
    BBCC*    next2;         /* entry chain in 2nd hash */
    Addr     array;         /* pointer to variable length array. If
			     * clo_simulate_cache is false, this points
			     * to same array for all BBCC of one BB
			     * (we only need instr. addresses for dump) */
};


/* the <number> of fn_node, file_node and obj_node are for compressed dumping
 * and a index into the dump boolean table and fn_info_table
 */

struct _fn_node {
    Char*      name;
    UInt       number;
    Context*   last_cxt; /* LRU info */
    Context*   pure_cxt; /* the context with only the function itself */
    file_node* file;     /* reverse mapping for 2nd hash */
    fn_node* next;

    Bool dump_before;
    Bool dump_after;
    Bool zero_before;
    Bool toggle_collect;
    Bool skip;
    Int  group;
    Int  separate_callers;
    Int  separate_recursions;
#if CT_ENABLE_DEBUG
    Int  verbosity; /* Stores old verbosity level while in function */
#endif
};

/* Quite arbitrary fixed hash sizes */

#define   N_OBJ_ENTRIES         47
#define  N_FILE_ENTRIES         53
#define    N_FN_ENTRIES         87
#define N_BBCC2_ENTRIES         37

struct _file_node {
   Char*      name;
   fn_node*   fns[N_FN_ENTRIES];
   UInt       number;
   obj_node*  obj;
   file_node* next;
};

/* If an object is dlopened multiple times, we hope that <name> is unique;
 * <start> and <offset> can change with each dlopen, and <start> is
 * zero when object is unmapped (possible at dump time).
 */
struct _obj_node {
   Char*      name;
   UInt       size;

   Addr       start;  /* Start address of text segment mapping */
   UInt       offset; /* Offset between symbol address and file offset */

   file_node* files[N_FILE_ENTRIES];
   UInt       number;
   obj_node*  next;
};

/* function and thread specific info.
 * Used in fn_info_table, using fn_node->number as index
 */
struct _fn_info {
    ULong call_counter;           /* how many times called at all */
    BBCC* bbccs[N_BBCC2_ENTRIES]; /* hash for fast deletion */
};


/* an entry in the callstack
 *
 * <nonskipped> is 0 if the function called is not skipped (usual case).
 * Otherwise, it is the last non-skipped BBCC. This one gets all
 * the calls to non-skipped functions and all costs in skipped 
 * instructions.
 */
struct _call_entry {
    jCC* jcc;           /* jCC for this call */
    fCC  fcc;           /* cost event counters at entering frame */
    Addr esp;           /* ESP at call time */
    BBCC* nonskipped;   /* see above */
    Context* cxt;       /* context before call */
    Int fn_sp;          /* function stack index before call */
};


/*
 * Thread context state while inside of a signal handler or in another thread.
 * As there's no scheduling among running signal handlers of one thread,
 * we only need a subset of a full thread state:
 * - event counter
 * - collect state
 * - last BB, last jump kind, last nonskipped BB
 * - callstack pointer for sanity checking and correct unwinding
 *   after exit
 */
struct _thread_cxtinfo {

    /* the signum of the handler, 0 for main thread context
     */
    Int sigNum;

    /* the call stack pointer at entering the signal handler */
    Int call_stack_bottom;

    fCC current;
    Bool collect_state;
    Context* current_cxt;

    /* last BB executed */
    Int  bbcc_jmpkind;
    BBCC* current_bbcc;
    BBCC* current_nonskipped;
};

/* The maximum number of simultaneous running signal handlers per thread.
 * This is used for the fixed size of stack of _thread_siginfo structs
 * in the thread state */
#define MAX_SIGHANDLERS 8

/* Thread State 
 *
 * This structure stores thread specific info while a thread is *not*
 * running. See function switch_thread() for save/restore on thread switch.
 *
 * If --separate-threads=no, BBCCs and JCCs can be shared by all threads, i.e.
 * only structures of thread 1 are used.
 * This involves variables fn_info_table, bbcc_table and jcc_table.
 */
struct _thread_info {

    /* resizable call stack */
    call_entry* stack;
    Int   stack_size;
    Int   stack_sp;

    /* event counters */
    fCC last;            /* counters at last dump */
    fCC discards;        /* events for discarded code */

    /* execution context */
    fn_node** fn_stack;
    fn_node** fn_stack_top;
    Int fn_stack_size;

     /* execution contexts (main thread + signal handlers) */
    thread_cxtinfo* cxt_stack[MAX_SIGHANDLERS];
    Int cxt_stackpointer; /* > 0 if a handler is running */
    fCC handler_sum;
    
    /* function active counts (used for fn-naming with recursion) */
    UInt* fn_active_array;
    Int fn_active_array_size;

    /* function info array: call counts & bbcc hash */
    fn_info** fn_info_table;
    Int       fn_info_table_size;

    /* thread dependend hashes: jcc, bbcc */
    Int jcc_table_size;
    Int jcc_table_entries;
    jCC **jcc_table;
    jCC *jcc_spontaneous;

    Int bbcc_table_size;
    Int bbcc_table_entries;
    BBCC ** bbcc_table;
};

/* Statistics */
struct _stat_info {
  fCC* current_fcc;

  ULong call_counter;
  ULong jump_counter;
  
  Int  distinct_objs;
  Int  distinct_files;
  Int  distinct_fns;
  Int  distinct_contexts;
  Int  distinct_bbs;
  Int  distinct_bbccs;
  Int  distinct_instrs;
  Int  distinct_jccs;
};


/*------------------------------------------------------------*/
/*--- Functions                                            ---*/
/*------------------------------------------------------------*/

/* from ct_sim.c */
void SK_(cachesim_init)(void);
void SK_(cachesim_getdesc)(Char*);
Int  SK_(cachesim_I1_doref)(Addr a, UChar size);
Int  SK_(cachesim_D1_doref)(Addr a, UChar size);
Int  SK_(cachesim_L2_doref)(Addr a, UChar size);

/* from ct_main.c */
void SK_(assert_fn_number_valid)(Int); /* TODO: move! */
call_entry* SK_(get_callentry)(Int);
ThreadId SK_(get_current_tid)(void);
Context* SK_(get_current_cxt)(void);
BBCC*    SK_(get_current_bbcc)(void);
Int      SK_(get_current_sp)(void);
Bool SK_(get_debug_info)(Addr, Char filename[FILENAME_LEN],
			 Char fn_name[FN_NAME_LEN], Int*, SegInfo**);
void SK_(set_instrument_state)(Char*,Bool);
void SK_(dump_profile)(Char* trigger,Bool only_current_thread);
void SK_(zero_ccs)(Bool only_current_thread);
stat_info* SK_(get_statistics)(void);
Int SK_(get_dump_counter)(void);
thread_info** SK_(get_threads)(void);
void SK_(switch_thread)(ThreadId);
void SK_(init_fcc)(fCC*);
void SK_(copy_fcc)(fCC*,fCC*);
Bool SK_(add_diff_fcc)(fCC*,fCC*,fCC*);
Int  SK_(sprint_fcc)(Char* buf, fCC* fcc);

/* from ct_clo.c */
void SK_(set_clo_defaults)(void);
void SK_(update_fn_config)(fn_node*);

/* from ct_command.c */
void SK_(init_command)(Char* dir, Char* dumps);
void SK_(check_command)(void);
void SK_(finish_command)(void);


/*------------------------------------------------------------*/
/*--- Exported global variables                            ---*/
/*------------------------------------------------------------*/

/* Number of basic blocks executed */
extern ULong SK_(bb_executions);

/* Function active counter array, indexed by function number */
extern UInt* SK_(fn_active_array);

extern Bool SK_(instrument_state);


/*------------------------------------------------------------*/
/*--- Debugging                                            ---*/
/*------------------------------------------------------------*/

#if CT_ENABLE_DEBUG

extern Int   clo_ct_verbose;
extern ULong clo_ct_verbose_start;

#define CT_DEBUGIF(x) \
  if ((clo_ct_verbose>x) && (SK_(bb_executions)>=clo_ct_verbose_start))

#define CT_DEBUG(x,format,args...)    \
    CT_DEBUGIF(x) {                   \
      SK_(print_bbno)();	      \
      VG_(printf)(format,##args);     \
    }

#define CT_ASSERT(cond)               \
    if (!(cond)) {                    \
      SK_(print_context)();           \
      sk_assert(cond);                \
     }

#else
#define CT_DEBUGIF(x) if (0)
#define CT_DEBUG(x,y) {}
#define CT_ASSERT(cond) sk_assert(cond);
#endif

/* from ct_debug.c */
#if CT_ENABLE_DEBUG
void SK_(print_bbno)(void);
void SK_(print_context)(void);
void SK_(print_jcc)(int s, jCC* jcc);
void SK_(print_bbcc)(int s, BBCC* bbcc);
void SK_(print_bbcc_fn)(BBCC* bbcc);
void SK_(print_fcc)(int s, fCC* fcc);
void SK_(print_bb)(int s, BB* bb);
void SK_(print_cxt)(int s, Context* cxt, int rec_index);
void SK_(print_short_jcc)(jCC* jcc);
void SK_(print_stackentry)(int s, int sp);
void SK_(print_addr)(Addr addr);
void SK_(print_addr_ln)(Addr addr);
#endif


#endif /* CT_INCLUDE */
